# -*- coding:utf-8 -*-

import os
from pprint import pprint
from delver import Crawler


def scraping_steam_specials():
    c = Crawler()
    c.logging = True
    c.useragent = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
    c.random_timeout = (0, 5)
    c.open('http://store.steampowered.com/search/?specials=1')
    titles, discounts, final_prices = [], [], []
    while c.links(filters={
        'class': 'pagebtn',
        'text': '>'
    }):
        c.open(c.current_results[0])
        titles.extend(
            c.xpath("//div/span[@class='title']/text()")
        )
        discounts.extend(
            c.xpath("//div[contains(@class, 'search_discount')]/span/text()")
        )
        final_prices.extend(
            c.xpath("//div[contains(@class, 'discounted')]//text()[2]").strip()
        )

    all_results = {
        row[0]: {
            'discount': row[1],
            'final_price': row[2]
        } for row in zip(titles, discounts, final_prices)}
    pprint(all_results)


def scraping_movies_table():
    c = Crawler()
    c.logging = True
    c.useragent = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
    c.open("http://www.boxofficemojo.com/daily/")
    pprint(c.tables())


def user_login():
    c = Crawler()
    c.useragent = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/60.0.3112.90 Safari/537.36"
    )
    c.random_timeout = (0, 5)
    c.open('http://testing-ground.scraping.pro/login')
    forms = c.forms()
    if forms:
        login_form = forms[0]
        login_form.fields = {
            'usr': 'admin',
            'pwd': '12345'
        }
        c.submit(login_form)
        success_check = c.submit_check(
            login_form,
            phrase='WELCOME :)',
            status_codes=[200]
        )
        print(success_check)


class OnePunchManDownloader:
    """Downloads One Punch Man free manga chapers to local directories.
    Uses one main thread for scraper with random timeout.
    Uses 20 threads just for image downloads.
    """
    def __init__(self):
        self._target_directory = 'one_punch_man'
        self._start_url = "http://m.mangafox.me/manga/onepunch_man_one/"
        self.crawler = Crawler()
        self.crawler.random_timeout = (0, 5)
        self.crawler.useragent = "Googlebot-Image/1.0"

    def run(self):
        self.crawler.open(self._start_url)
        for link in self.crawler.links(filters={'text': 'Ch '}, match='IN'):
            self.download_images(link)

    def download_images(self, link):
        target_path = '{}/{}'.format(self._target_directory, link.split('/')[-2])
        full_chapter_url = link.replace('/manga/', '/roll_manga/')
        self.crawler.open(full_chapter_url)
        images = self.crawler.xpath("//img[@class='reader-page']/@data-original")
        os.makedirs(target_path, exist_ok=True)
        self.crawler.download_files(target_path, files=images, workers=20)


def one_punch_downloader():
    downloader = OnePunchManDownloader()
    downloader.run()



# -*- coding: utf-8 -*-

from functools import wraps

from .exceptions import HistoryError


def with_history(func):
    """Ensures that history is set to true for methods that require history"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if args[0]._history:
            return func(*args, **kwargs)
        else:
            raise HistoryError("Crawler history is off.")
    return wrapper



# -*- coding: utf-8 -*-

from lxml.html import HtmlElement

from .helpers import table_to_dict, filter_element


class Scraper:
    """Scraping methods for `Crawler`
    """

    def __init__(self, *args, **kwargs):
        self.current_results = []

    def css(self, selector):
        """Wraps lxml parser css method"""
        results = self._parser.css(selector)
        if not isinstance(results, list):
            results = [results]
        self.current_results = ResultsList(results)
        return self.current_results

    def xpath(self, path):
        """Wraps lxml parser xpath method"""
        results = self._parser.xpath(path)
        if not isinstance(results, list):
            results = [results]
        self.current_results = ResultsList(results)
        return self.current_results

    def regexp(self, selectors=None):
        """Not implemented due the poor performance."""
        raise NotImplementedError

    def title(self):
        """Scrapes website titles

        :return: list
        """
        return self._parser.xpath('//title/text()')

    def links(self, tags=None, filters=None, match='EQUAL'):
        """Find all links on current page using given criteria

        Usage::

            >>> c = Crawler()
            >>> c.open('https://httpbin.org/links/10/0')
            <Response [200]>
            >>> links = c.links(
            ...     tags = ('style', 'link', 'script', 'a'),
            ...     filters = {
            ...         'text': '7'
            ...     },
            ...     match='NOT_EQUAL'
            ... )
            >>> len(links)
            8

        :param tags: allowed html tags (like 'style', 'link', 'script', 'a')
        :param filters: dictionary of filters, possible values: id, text, title, class
        :param match: type of matching, possible values: 'IN', 'NOT_IN', 'EQUAL', 'NOT_EQUAL'
        :return:
        """
        self.current_results = ResultsList(
            list(
                self.current_parser().find_links(
                    tags,
                    filters,
                    match
                ).keys()
            )
        )
        return self.current_results

    def images(self, filters=None, match='EQUAL'):
        """ Scraping images using filtering

        :param tags: allowed html tags (like 'style', 'link', 'script', 'a')
        :param filters: dictionary of filters, possible values: id, text, title, class
        :param match: type of matching, possible values: 'IN', 'NOT_IN', 'EQUAL', 'NOT_EQUAL'
        :return:
        """
        images = []
        filters = filters or {}
        for image in self._parser.xpath('//img'):
            src = image.attrib.get('src')
            if src:
                matched = filter_element(
                    image,
                    filters=filters,
                    match=match,
                    custom_attrs=['alt', 'src']
                )
                if matched:
                    images.append(src)
        self.current_results = ResultsList(images)
        return self.current_results

    def tables(self):
        """Scrapes tables to list of dicts. Works only with simple flat tables with <th> headers.

        :return: list of defaultdicts
        """
        return [
            table_to_dict(table)
            for table in self._parser.xpath('//table')
        ]


class ResultsList:

    __slots__ = ['results']

    def __init__(self, results):
        self.results = results or []

    def __getattr__(self, item):
        return getattr(self.results, item)

    def __len__(self):
        return self.results.__len__()

    def __getitem__(self, item):
        return self.results.__getitem__(item)

    def filter(self, tags=None, filters=None, match='EQUAL', custom_attrs=None):
        """Filters results list. Item in a list should be instances of `HtmlElement`.
        This method is to narrow result list of scraped elements. It's recommended to use this
        method on relatively small amount of results. It can be done by appropriate use of
        xpath and css selectors.

        :param tags: allowed html tags (like 'style', 'link', 'script', 'a')
        :param filters: dictionary of filters, possible values: id, text, title, class
        :param match: type of matching, possible values: 'IN', 'NOT_IN', 'EQUAL', 'NOT_EQUAL'
        :param custom_attrs: custom attrs could be added to filters, like `src, alt` for example
        :return: list of dicts
        """
        tags = tags or []
        filters = filters or {}
        filtered_results = []
        for item in self.results:
            if isinstance(item, HtmlElement):
                matched = filter_element(
                    item,
                    tags=tags,
                    filters=filters,
                    match=match,
                    custom_attrs=custom_attrs
                )
                if matched:
                    filtered_results.append(matched)
        return filtered_results

    def strip(self):
        return [result.strip() for result in self.results]



# -*- coding: utf-8 -*-


class Useragent:

    def __get__(self, instance, owner):
        return instance.__dict__['_useragent']

    def __set__(self, instance, value):
        """Sets useragent. Useragent set in that way is used for all crawler requests
        unless it is overridden in request kwargs or changed.
        """
        instance.__dict__['_useragent'] = value
        instance.__dict__['_headers']['user-agent'] = value


class Proxy:

    def __get__(self, instance, owner):
        return instance.__dict__['_proxy']

    def __set__(self, instance, value):
        """Sets proxy. Proxy set in that way is used for all crawler requests
        unless it is overridden in request kwargs or changed.
        """
        instance.__dict__['_proxy'] = {
            'http': 'http://{}'.format(value),
            'https': 'https://{}'.format(value)
        }


class Headers:

    def __get__(self, instance, owner):
        return instance.__dict__['_headers']

    def __set__(self, instance, value):
        """Sets headers. Headers set in that way are used for all crawler requests
        unless it is overridden in request kwargs or changed.
        """
        instance.__dict__['_headers'] = value



# -*- coding: utf-8 -*-

import io

from urllib.parse import urljoin, urlparse

from lxml.html import (
    CheckboxGroup,
    CheckboxValues,
    SelectElement,
    RadioGroup
)

PARENT_METHODS = ['form_values', 'action', 'method']


def prepare_field(field):
    data = {
        'value': field.value,
        'tag': field.tag if hasattr(field, 'type') else None,
        'type': field.type if hasattr(field, 'type') else None,
    }
    if isinstance(field, CheckboxValues):
        data['value_options'] = field.group.value_options
    elif isinstance(field, CheckboxGroup):
        data.update({
            'value': list(field.value),
            'value_options': field.value_options
        })
    elif isinstance(field, SelectElement):
        data.update({
            'value': list(field.value) if field.multiple else field.value,
            'value_options': field.value_options
        })
    elif isinstance(field, RadioGroup):
        data['value_options'] = field.value_options
    return data


class FormWrapper:
    """Wrapper for ``lxml.html.FormElement``.

    Represents a <form> element.
    Basically, it's FormElement with added:
    - Ability to submit itself - method ``submit``
    - Success/Fail check after submit form - method ``check``
    - Readable representation of fields - property ``fields``
    - Prepare full action url - method ``action_url``
    - Checking if there are specific fields in form - method ``has_fields``
    - Simplified api to access form attributes like ``id``, ``name``

    """

    def __init__(self, lxml_form, session=None, url=None):
        """FormWrapper initialization

        :param lxml_form: class::`FormElement <FormElement>` object
        :param session: session object required only if we want to submit form
        :param url: current base url used to assemble full action url
        """
        self._lxml_form = lxml_form
        self._fields = {}
        self._files = {}
        self._session = session
        self._result = None
        self._url = url or None

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, value):
        self._result = value

    @property
    def files(self):
        return self._files

    @files.getter
    def files(self):
        return self._files

    @files.setter
    def files(self, value):
        self._files = value

    @property
    def fields(self):
        """Readable representation of fields as a ``dict``.

        :return: dict
        """
        for name, field in dict(self._lxml_form.inputs).items():
            self._fields[name] = prepare_field(field)
        return self._fields

    @fields.setter
    def fields(self, values):
        """Sets form fields (inputs) values.

        :param values: dict with form
        :return:
        """
        text_values = {}
        for name, val in values.items():
            if isinstance(val, io.IOBase):
                self._files[name] = val
            else:
                text_values[name] = val

        self._lxml_form.fields = text_values

    def id(self):
        """Returns form id attribute

        :return: str
        """
        return self._lxml_form.attrib.get('id')

    def name(self):
        """Returns form name attribute

        :return: str
        """
        return self._lxml_form.attrib.get('name')

    def action_url(self):
        """Assembles full action url

        :return: str
        """
        action_url = self.action
        return urljoin(
            self._url if urlparse(action_url).scheme else '',
            self.action
        )

    def has_fields(self, fields):
        """Return ``True`` if all fields are present

        :return: bool
        """
        for field in fields:
            if field not in dict(self._lxml_form.inputs):
                return False
        return True

    def __getattr__(self, item):
        """Wraps some chosen methods (PARENT_METHODS) from ``lxml.html.FormElement``

        :param item: method name
        :return: value of the attribute
        """
        if item in PARENT_METHODS:
            return getattr(self._lxml_form, item)
        return getattr(self, item)

    def append_extra_values(self, values, extra_values):
        """ Extends form values by adding extra values.

        :param values: current form fields values
        :param extra_values: additional values extending current values
        :return: None
        """
        if extra_values:
            if hasattr(extra_values, 'items'):
                extra_values = extra_values.items()
            values.extend(extra_values)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding:utf-8 -*-

import os
import json
import logging.config


def setup_logging(default_path='logging.json', default_level=logging.ERROR):
    path = default_path
    if os.path.exists(path):
        with open(path) as f:
            config = json.load(f)
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)



# -*- coding: utf-8 -*-

import asyncio
from concurrent.futures import ThreadPoolExecutor

import requests

from .helpers import ForcedInteger


class Proxy:
    """Proxy object

    Wraps proxy ip address like "110.136.228.250:80" to object. Provides testing methods.

    """
    timeout = ForcedInteger('timeout')

    def __init__(self, proxy, _type=None, test_url=None, timeout=None):
        """Proxy initialization

        :param proxy: ip address like "110.136.228.250:80"
        :param _type: type of proxy
        :param test_url: url used during proxy testing
        :param timeout: max number of seconds to complete request
        """
        self.address = proxy
        self.working = True
        self._type = _type
        self._test_url = test_url or 'https://httpbin.org/ip'
        self._timeout = timeout or 15
        self._errors = 0

    def test(self):
        """ Test if proxy works.

        Loads custom page through proxy and checks if proxy ip address is in the response.

        :return: bool test result
        """
        request = self.proxy_request()
        self.working = (
            request.status_code == 200
            and self.address in request.json().get('origin', '')
        )
        self._errors += bool(self.working)
        return self.working

    def proxy_request(self):
        return requests.get(
            self._test_url,
            proxies={
                'http': f'http://{self.address}',
                'https': f'https://{self.address}'
            },
            timeout=self._timeout
        )

    def __repr__(self):
        return f'<Proxy(address={self.address}, working={self.working})>'


class ProxyPool:
    """Pool for handling proxies list.

    Allows to load, check and in generally manage proxies.

    Usage::

    >>> proxies = [
    ...     "110.136.228.250:80",
    ...     "166.78.156.247:80",
    ...     "173.234.216.40:21320"
    ... ]
    >>> proxy_pool = ProxyPool()
    >>> proxy_pool.load_proxies(proxies, test=False)
    >>> len(list(proxy_pool.working()))
    3

    .. todo:: Add loading from file and adding, removing single items
    """
    def __init__(self, workers=10):
        """ ProxyPool initialization

        :param workers: max workers number for executor
        """
        self._proxies = []
        self._executor = ThreadPoolExecutor(max_workers=workers)
        self._loop = None

    def __getattr__(self, name):
        """Magically extends ProxyPool methods of list methods like append, insert, sort etc.

        :param name: attribute name
        :return: attribute of list
        """
        return getattr(self._proxies, name)

    def load_proxies(self, proxies, test=False):
        """ Loads proxies. Extends pool <Proxy> list.

        :param proxies:
        :param test:
        :return:
        """
        self._loop = asyncio.get_event_loop()
        self._proxies.extend(
            self._loop.run_until_complete(self.tasks_from_list(proxies, test=test))
        )
        self._loop.close()

    async def tasks_from_list(self, proxies, test=False):
        """ Prepares list of tasks and runs them in executor.

        :param proxies: list of proxies
        :param test: test if works or not
        :return: list of <Proxy> objects
        """
        tasks = [
            self._loop.run_in_executor(self._executor, self.load_proxy, address, test)
            for address in proxies
        ]
        completed, pending = await asyncio.wait(tasks)
        return [task.result() for task in completed]

    async def tasks_from_file(self, proxies, test=False):
        raise NotImplementedError

    def working(self):
        return WorkingProxyList(proxies=self._proxies)

    def __iter__(self):
        return self._proxies.__iter__()

    def __len__(self):
        return len(self._proxies)

    def __getitem__(self, item):
        return self._proxies[item]

    def load_proxy(self, address, test=False):
        """ Makes <Proxy> object from ip address.

        :param address: proxy address "110.136.228.250:80"
        :param test: test if works or not
        :return: <Proxy> object
        """
        try:
            proxy = Proxy(address)
            if test:
                proxy.test()
        except Exception as err:
            proxy.working = False
            print(repr(err))
        finally:
            return proxy


class ProxyList:

    def __init__(self, proxies=None):
        self._proxies = proxies or []


class WorkingProxyList(ProxyList):

    def __iter__(self):
        return (proxy for proxy in self._proxies if proxy.working)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-

import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from random import randrange

from collections import namedtuple, deque
from copy import deepcopy
from urllib.parse import urljoin, urlparse

import requests

from .decorators import with_history
from .exceptions import CrawlerError
from .helpers import ForcedInteger
from .parser import HtmlParser
from .scraper import Scraper
from .descriptors import (
    Useragent,
    Proxy,
    Headers
)

PARSERS = {
    'text/html': HtmlParser,
    'text/plain': HtmlParser,
    'text/json': HtmlParser,
    'application/xml': HtmlParser,
    'application/json': HtmlParser
}


class Crawler(Scraper):
    """Browser mimicking object. Mostly wrapper on Requests and Lxml libraries.

    :param history: (optional) bool, turns off/on history usage in Crawler
    :param max_history: (optional) int, max items held in history
    :param absolute_links: (optional) bool, makes always all links absolute


    Features:

    - To some extent, acts like a browser

    - Allows visiting pages, form posting, content scraping, cookie handling etc.

    - Wraps ``requests.Session()``


    Simple usage::

        >>> c = Crawler()
        >>> response = c.open('https://httpbin.org/html')
        >>> response.status_code
        200


    Form submit::

        >>> c = Crawler()
        >>> response = c.open('https://httpbin.org/forms/post')
        >>> forms = c.forms()

        Filling up fields values:
        >>> form = forms[0]
        >>> form.fields = {
        ...    'custname': 'Ruben Rybnik',
        ...    'custemail': 'ruben.rybnik@fakemail.com',
        ...    'size': 'medium',
        ...    'topping': ['bacon', 'cheese'],
        ...    'custtel': '+48606505888'
        ... }
        >>> submit_result = c.submit(form)
        >>> submit_result.status_code
        200

        Checking if form post ended with success:
        >>> c.submit_check(
        ...    form,
        ...    phrase="Ruben Rybnik",
        ...    url='https://httpbin.org/post',
        ...    status_codes=[200])
        True


    Form file upload::

        >>> c = Crawler()
        >>> c.open('http://cgi-lib.berkeley.edu/ex/fup.html')
        <Response [200]>
        >>> forms = c.forms()
        >>> upload_form = forms[0]
        >>> upload_form.fields = {
        ...    'note': 'Text file with quote',
        ...    'upfile': open('test/test_file.txt', 'r')
        ... }
        >>> c.submit(upload_form, action='http://cgi-lib.berkeley.edu/ex/fup.cgi')
        <Response [200]>
        >>> c.submit_check(
        ...    upload_form,
        ...    phrase="road is easy",
        ...    status_codes=[200]
        ... )
        True

    Cookies handling::

        >>> c = Crawler()
        >>> c.open('https://httpbin.org/cookies', cookies={
        ...     'cookie_1': '1000101000101010',
        ...     'cookie_2': 'ABABHDBSBAJSLLWO',
        ... })
        <Response [200]>

    Find links::

        >>> c = Crawler()
        >>> c.open('https://httpbin.org/links/10/0')
        <Response [200]>

        Links can be filtered by some html tags and filters
        like: id, text, title and class:
        >>> links = c.links(
        ...     tags = ('style', 'link', 'script', 'a'),
        ...     filters = {
        ...         'text': '7'
        ...     },
        ...     match='NOT_EQUAL'
        ... )
        >>> len(links)
        8

    Find images::

        >>> c = Crawler()
        >>> c.open('https://www.python.org/')
        <Response [200]>

        First image path with 'python-logo' in string:
        >>> next(
        ...     image_path for image_path in c.images()
        ...     if 'python-logo' in image_path
        ... )
        'https://www.python.org/static/img/python-logo.png'

    Download file::

        >>> import os

        >>> c = Crawler()
        >>> local_file_path = c.download(
        ...     local_path='test',
        ...     url='https://httpbin.org/image/png',
        ...     name='test.png'
        ... )
        >>> os.path.isfile(local_file_path)
        True

    Download files list in parallel::

        >>> c = Crawler()
        >>> c.open('https://xkcd.com/')
        <Response [200]>
        >>> full_images_urls = [c.join_url(src) for src in c.images()]
        >>> downloaded_files = c.download_files('test', files=full_images_urls)
        >>> len(full_images_urls) == len(downloaded_files)
        True

    Traversing through history::

        >>> c = Crawler()
        >>> c.open('http://quotes.toscrape.com/')
        <Response [200]>
        >>> tags_links = c.links(filters={'class': 'tag'})
        >>> c.follow(tags_links[0])
        <Response [200]>
        >>> c.follow(tags_links[1])
        <Response [200]>
        >>> c.follow(tags_links[2])
        <Response [200]>
        >>> history = c.history()
        >>> c.back()
        >>> c.get_url() == history[-2].url
        True
    """

    useragent = Useragent()
    proxy = Proxy()
    headers = Headers()
    max_retries = ForcedInteger('max_retries')

    def __init__(self, history=True, max_history=5, absolute_links=True):
        """Crawler initialization

        :param history: bool, turns on/off history handling
        :param max_history: max items stored in flow
        :param absolute_links: globally make links absolute
        """
        super().__init__(
            history=history,
            max_history=max_history,
            absolute_links=absolute_links
        )
        self._session = requests.Session()
        self._history = history
        self._max_history = max_history
        self._flow = deque(maxlen=self._max_history)
        self._index = 0
        self._parser = None
        self._current_response = None
        self._absolute_links = absolute_links
        self._useragent = None
        self._headers = {}
        self._proxy = {}
        self._loop = None
        self._executor = None
        self._max_retries = 0
        self._retries = 0
        self._logging = False
        self._logger = None
        self._random_timeout = None

    @property
    def logging(self):
        return self._logging

    @logging.setter
    def logging(self, value):
        self._logging = value
        self._logger = logging.getLogger(__name__) if value else None

    @property
    def random_timeout(self):
        return self._random_timeout

    @random_timeout.setter
    def random_timeout(self, value):
        if isinstance(value, (list, tuple)):
            self._random_timeout = value
        else:
            raise TypeError('Expected list or tuple.')

    def fit_parser(self, response):
        """Fits parser according to response type.

        :param response: class::`Response <Response>` object
        :return: matched parser object like: class::`HtmlParser <HtmlParser>` object
        """
        content_type = response.headers.get('Content-type', '')
        for _type, parser in PARSERS.items():
            if _type in content_type:
                self._parser = PARSERS[_type](response, session=self._session)
                return self._parser
        if self._logging:
            self._logger.info("Couldn't fit parser for {}.".format(content_type))

    def handle_response(self):
        """Called after request. Make operations accordng to attributes settings."""
        if self._absolute_links:
            self._parser.make_links_absolute()
        if self._history:
            self._flow.append({'parser': deepcopy(self._parser)})

    def open(self, url, method='get', **kwargs):
        """Opens url. Wraps functionality of `Session` from `Requests` library.

        :param url: visiting url str
        :param method: 'get', 'post' etc. str
        :param kwargs: additional keywords like headers, cookies etc.
        :return: class::`Response <Response>` object
        """
        self._retries = 0
        self._current_response = None
        flow_len = len(self._flow)
        if flow_len < self._max_history:
            self._index = flow_len

        self.add_customized_kwargs(kwargs)

        while True:
            try:
                self._current_response = self._session.request(method, url, **kwargs)
                if self._random_timeout:
                    time.sleep(randrange(*self._random_timeout))
                if self._logging:
                    self._logger.info(
                        'Open method: {} request: url={}, status code={}, kwargs={}  '.format(
                            method.upper(),
                            url,
                            self._current_response.status_code,
                            kwargs
                        ))
            except requests.exceptions.ConnectionError:
                self._retries += 1
                time.sleep(self._retries)
                if self.logging:
                    self._logger.error(
                        'Failed, try {}, method: {} request: url={}, kwargs={}  '.format(
                            self._retries,
                            method.upper(),
                            url,
                            kwargs
                        ))
                if self._retries >= self._max_retries:
                    raise
                continue
            break

        if self._current_response and self.fit_parser(self._current_response):
            self.handle_response()
            if self._history:
                self._flow[self._index].update({'response': deepcopy(self._current_response)})
            return self._current_response

    def add_customized_kwargs(self, kwargs):
        """Adds request keyword arguments customized by setting `Crawler`
        attributes like proxy, useragent, headers. Arguments won't be passed
        if they are already set as `open` method kwargs.
        """
        if self._proxy and 'proxies' not in kwargs:
            kwargs.update({'proxies': self._proxy})
        if self._headers and 'headers' not in kwargs:
            kwargs.update({'headers': self._headers})

    def response(self):
        """Get current response."""
        return self._current_response

    def get_url(self):
        """Get URL of current document."""
        return self._current_response.url

    def join_url(self, url_path):
        """Returns absolute_url. Path joined with url_root."""
        return urljoin(
            self._current_response.url,
            url_path
        )

    @with_history
    def back(self, step=1):
        """Go back n steps in history, and return response object"""
        if self._index - step > 0:
            self._index -= step
            self._current_response = self._flow[self._index]['response']
        else:
            raise CrawlerError("Out of history boundaries")

    @with_history
    def forward(self, step=1):
        """Go forward n steps in history, and return response object"""
        if self._index + step < self._max_history:
            self._index += step
            self._current_response = self._flow[self._index]['response']
        else:
            raise CrawlerError("Out of history boundaries")

    def follow(self, url, method='get',  **kwargs):
        """Follows url"""
        self.add_customized_kwargs(kwargs)
        return self.open(self.join_url(url), method, **kwargs)

    @with_history
    def flow(self):
        """Return flow"""
        return self._flow

    def clear(self):
        """Clears all flow, session, headers etc."""
        self._flow.clear()
        self._index = 0
        self._session.cookies.clear()
        self._headers = {}
        self._proxy = {}

    @with_history
    def history(self):
        """Return urls history and status codes"""
        Visit = namedtuple('Visited', 'url method response')
        return [
            Visit(
                history['response'].url,
                history['response'].request.method,
                history['response'].status_code
            )
            for history in self._flow
        ]

    def request_history(self):
        """Returns current request history (like list of redirects to finally accomplish request)
        """
        return self._current_response.history

    @property
    def cookies(self):
        """Wraps `RequestsCookieJar` object from requests library.

        :return: `RequestsCookieJar` object
        """
        return self._current_response.cookies

    def current_parser(self):
        """Return parser associated with current flow item.

        :return: matched parser object like: class::`HtmlParser <HtmlParser>` object
        """
        return self._flow[self._index]['parser']

    def forms(self, filters=None):
        """Return iterable over forms. Doesn't find javascript forms yet (but will be).

            example_filters = {
                'id': 'searchbox',
                'name': 'name,
                'action': 'action',
                'has_fields': ['field1', 'field2'],
            }

        Usage::

            >>> c = Crawler()
            >>> response = c.open('http://cgi-lib.berkeley.edu/ex/fup.html')
            >>> forms = c.forms()
            >>> forms[0].fields['note'].get('tag')
            'input'
            """

        filters = filters or {}
        if self._history:
            return self.current_parser().find_forms(filters)
        return self._parser.find_forms(filters)

    def submit(self, form=None, action=None, data=None):
        """Submits form

        :param form: `FormWrapper` object
        :param action: custom action url
        :param data: additional custom values to submit
        :return: submit result
        """
        if form:
            action = action or form.action_url()
            values = form.form_values()
            form.append_extra_values(values, data)
            form.result = self.open(
                action,
                form.method,
                data=values,
                files=form.files,
            )
        else:
            self.direct_submit(url=action, data=data)
        return self._current_response

    def direct_submit(self, url=None, data=None):
        """Direct submit. Used when quick post to form is needed or if there are no forms found
        by the parser.

        Usage::

            >>> data = {'name': 'Piccolo'}
            >>> c = Crawler()
            >>> result = c.submit(action='https://httpbin.org/post', data=data)
            >>> result.status_code
            200

        :param url: submit url, form action url, str
        :param data: submit parameters, dict
        :return: class::`Response <Response>` object
        """
        current_url = None
        if self._current_response:
            current_url = self._current_response.url
        return self.open(
            url or current_url,
            method='post',
            data=data or {}
        )

    def submit_check(self, form, phrase=None, url=None, status_codes=None):
        """Checks if success conditions of form submit are met

        :param form: `FormWrapper` object
        :param phrase: expected phrase in text
        :param url: expected url
        :param status_codes: list of expected status codes
        :return: bool
        """
        return all([
            phrase in form.result.text if phrase else True,
            form.result.url == url if url else True,
            form.result.status_code in status_codes if status_codes else True
        ])

    def encoding(self):
        """Returns current respose encoding."""
        return self._flow[self._index].encoding

    def download(self, local_path=None, url=None, name=None):
        file_name = name or os.path.split(urlparse(url).path)[-1]
        if file_name:
            download_path = os.path.join(local_path, file_name)
            with open(download_path, 'wb') as f:
                f.write(self._session.get(url).content)
            return download_path

    def download_files(self, local_path, files=None, workers=10):
        """Download list of files in parallel.

        :param workers: number of threads
        :param local_path: download path
        :param files: list of files
        :return: list with downloaded files paths
        """
        files = files or []
        results = []
        with ThreadPoolExecutor(max_workers=workers) as executor:
            for future in as_completed(
                    executor.submit(self.download, local_path, file)
                    for file in files
            ):
                results.append(future.result())

        return results


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding:utf-8 -*-

import os
import doctest
import shutil

from delver import (
    crawler,
    forms,
    helpers,
    parser,
    proxies
)

if __name__ == "__main__":
    os.makedirs('test', exist_ok=True)
    with open('test/test_file.txt', 'wb') as f:
        f.write(b"If the road is easy, you're likely going the wrong way..")
    doctest.testmod(crawler)
    doctest.testmod(forms)
    doctest.testmod(helpers)
    doctest.testmod(parser)
    doctest.testmod(proxies)
    shutil.rmtree('test', ignore_errors=True)



# -*- coding:utf-8 -*-

import os
import shutil
import unittest
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

from requests.exceptions import ConnectionError

from .crawler import Crawler
from .exceptions import CrawlerError
from .proxies import ProxyPool


class TestAll(unittest.TestCase):

    def setUp(self):
        self.urls = {
            'SIMPLE_HTML': 'https://httpbin.org/html',
            'SIMPLE_TABLE': 'https://www.w3schools.com/html/html_tables.asp',
            'USER_AGENT': 'https://httpbin.org/user-agent',
            'POST': 'https://httpbin.org/post',
            'COMPLEX_HTML': 'https://www.nytimes.com/',
            'FORM': 'https://httpbin.org/forms/post',
            'FORM2': 'https://eoryginalne.pl/kontakt',
            'GOOGLE': 'https://www.google.pl/',
            'LINKS': 'https://httpbin.org/links/10/0',
            'ROBOTS': 'https://httpbin.org/robots.txt',
            'IMAGE': 'https://httpbin.org/image/png',
            'FILE_UPLOAD': 'http://cgi-lib.berkeley.edu/ex/fup.html',
            'XML': 'https://httpbin.org/xml',
            'REDIRECT': 'https://httpbin.org/redirect/1',
            'REDIRECT_2_TIMES': 'https://httpbin.org/redirect/2',
            'COOKIES': 'https://httpbin.org/cookies',
            'BASIC_AUTH': 'https://httpbin.org/basic-auth/user/passwd',
            'GZIP': 'https://httpbin.org/gzip',
            'REST': 'https://jsonplaceholder.typicode.com/posts',
            'SCRAPING_QUOTES': 'http://quotes.toscrape.com/',
            'SCRAPING_BOOKSTORE': 'http://books.toscrape.com/',
            'GAZETA': 'http://www.gazeta.pl/0,0.html',
            'XKCD': 'https://xkcd.com/',
            'PYTHON': 'https://www.python.org/',
            'W3': 'https://www.w3schools.com/'
        }
        self.urls_list = [
            "https://www.nytimes.com/",
            "http://www.alexa.com/",
            "https://www.mozilla.org/en-US/",
            "https://www.amazon.com/",
            "https://www.walmart.com/",
            "http://www.biedronka.pl/pl",
            "https://pl.aliexpress.com/",
            "https://allegro.pl/",
            "https://httpbin.org/html",
            "https://github.com/"
        ]
        self.test_dir = os.path.join(os.getcwd(), 'test')
        self.upload_file = os.path.join(self.test_dir, 'upload.txt')
        os.makedirs(self.test_dir, exist_ok=True)

        with open(self.upload_file, 'wb') as f:
            f.write(b"If the road is easy, you're likely going the wrong way..")

    def tearDown(self):
        shutil.rmtree(self.test_dir, ignore_errors=True)

    def test_simple_http_page(self):
        c = Crawler()
        response = c.open(self.urls['SIMPLE_HTML'])
        self.assertEqual(response.status_code, 200)

    def test_find_links(self):
        c = Crawler()
        c.open(self.urls['GAZETA'])
        links = c.links(filters={
            'id': 'LinkArea:MT'
            }, match='EQUAL'
        )
        self.assertTrue(links)

    def test_parse_form(self):
        c = Crawler()
        response = c.open(self.urls['FORM2'])
        self.assertEqual(response.status_code, 200)
        forms = c.forms(filters={'id': 'searchbox'})
        search_form = forms[0]
        search_form.fields = {
            'search_query': 'cute kittens'
        }
        self.assertEqual(search_form.fields['search_query'].get('value'), 'cute kittens')

    def test_submit_form(self):
        c = Crawler()
        response = c.open(self.urls['FORM'])
        self.assertEqual(response.status_code, 200)
        forms = c.forms()
        form = forms[0]
        form.fields = {
            'custname': 'aaa',
            'delivery': '',
            'custemail': 'test@email.com',
            'comments': '',
            'size': 'medium',
            'topping': ['bacon', 'cheese'],
            'custtel': '+48606505888'
        }
        c.submit(form, data={'extra_value': "I am your father."})
        success = c.submit_check(
            form,
            phrase="I am your father.",
            url=self.urls['POST'],
            status_codes=[200]
        )
        self.assertEqual(success, True)

    def test_response_history(self):
        c = Crawler()
        response = c.open(self.urls['REDIRECT_2_TIMES'])
        self.assertEqual(len(response.history), 2)

    def test_crawler_history(self):
        c = Crawler()
        c.open(self.urls['SIMPLE_HTML'])
        c.open(self.urls['LINKS'])
        history = c.history()
        self.assertEqual(history[0].url, self.urls['SIMPLE_HTML'])
        self.assertEqual(history[1].url, self.urls['LINKS'])

    def test_crawler_flow(self):
        c = Crawler()
        c.open(self.urls['SIMPLE_HTML'])
        c.open(self.urls['LINKS'])
        self.assertEqual(len(c.flow()), 2)

    def test_crawler_xpath(self):
        c = Crawler()
        c.open(self.urls['SIMPLE_HTML'])
        p_text = c.xpath('//p/text()')
        self.assertGreaterEqual(len(p_text), 1)

    def test_crawler_xpath_filter(self):
        c = Crawler()
        c.open(self.urls['W3'])
        filtered_results = c.xpath('//p').filter(filters={'class': 'w3-xlarge'})
        self.assertEqual(filtered_results[0]['class'], 'w3-xlarge')

    def test_crawler_css(self):
        c = Crawler()
        c.open(self.urls['SIMPLE_HTML'])
        p_text = c.css('div')
        self.assertGreaterEqual(len(p_text), 1)

    def test_crawler_scraper_methods(self):
        c = Crawler()
        c.open(self.urls['SIMPLE_TABLE'])
        self.assertTrue(c.tables())
        self.assertTrue(c.title())
        self.assertTrue(c.images())

    def test_crawler_back_forward_navigation(self):
        c = Crawler()
        c.open(self.urls['SCRAPING_QUOTES'])
        tags_links = c.links(filters={'class': 'tag'})
        for link in tags_links:
            c.follow(link)
        history = c.history()
        c.back()
        self.assertEqual(c.get_url(), history[-2].url)
        c.back()
        self.assertEqual(c.get_url(), history[-3].url)
        c.forward()
        self.assertEqual(c.get_url(), history[-2].url)
        self.assertRaises(CrawlerError, c.forward, step=5)

    def test_crawler_clear_flow(self):
        c = Crawler()
        c.open(self.urls['SCRAPING_QUOTES'])
        tags_links = c.links(filters={'class': 'tag'})
        for link in tags_links:
            c.follow(link)
        self.assertTrue(c.history())
        c.clear()
        self.assertFalse(c.history())

    def test_crawler_cookies_handling(self):
        c = Crawler()
        c.open(self.urls['COOKIES'], cookies={
            'cookie_1': '1000101000101010',
            'cookie_2': 'ABABHDBSBAJSLLWO',
        })
        response = c.response().json()
        self.assertIn('cookie_1', response.get('cookies'), {})

    def test_crawler_useragent_set(self):
        c = Crawler()
        c.useragent = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
        json_response = c.open(self.urls['USER_AGENT']).json()
        self.assertEqual(c.useragent, json_response['user-agent'])

    def test_crawler_overridden_useragent(self):
        c = Crawler()
        c.useragent = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"
        one_time_useragent = 'Godzilla'
        json_response = c.open(
            self.urls['USER_AGENT'],
            headers={'user-agent': one_time_useragent}
        ).json()
        self.assertEqual(one_time_useragent, json_response['user-agent'])

    def test_download_file(self):
        c = Crawler()
        file_name = 'test.png'
        c.download(
            self.test_dir,
            self.urls['IMAGE'],
            name=file_name
        )
        self.assertTrue(os.path.isfile(os.path.join(self.test_dir, file_name)))

    def test_download_images_list(self):
        """ Download list of images in parallel.
        """
        c = Crawler()
        c.open(self.urls['XKCD'])
        full_images_urls = [c.join_url(src) for src in c.images()]
        downloaded_files = c.download_files(self.test_dir, files=full_images_urls)
        self.assertEqual(len(full_images_urls), len(downloaded_files))

    def test_form_file_upload(self):
        """Uploading a file through """
        c = Crawler()
        c.open(self.urls['FILE_UPLOAD'])
        forms = c.forms()
        upload_form = forms[0]
        upload_form.fields = {
            'note': 'Towel cat picture',
            'upfile': open(self.upload_file, 'r')
        }
        c.submit(upload_form, action='http://cgi-lib.berkeley.edu/ex/fup.cgi')
        success = c.submit_check(
            upload_form,
            phrase="road is easy",
            status_codes=[200]
        )
        self.assertTrue(success)

    def test_proxy_pool(self):
        proxies = [
            "117.143.109.159:80",
            "117.143.109.163:80",
            "162.243.108.161:3128",
            "195.14.242.39:80",
            "202.57.129.228:8080",
            "212.124.171.144:80",
            "216.249.79.140:21320",
            "220.130.34.177:80"
            "52.10.247.166:80",
            "77.51.16.170:80",
        ]
        proxy_pool = ProxyPool()
        proxy_pool.load_proxies(proxies, test=False)
        working = list(proxy_pool.working())
        proxy_pool_len = len(proxy_pool)
        self.assertEqual(len(proxies), proxy_pool_len)
        self.assertEqual(len(working), proxy_pool_len)

    def test_run_crawler_in_threads(self):
        c = Crawler()
        with ThreadPoolExecutor(max_workers=4) as executor:
            results = executor.map(c.open, self.urls_list)
        self.assertEqual(len(list(results)), 10)

    def test_run_crawler_in_processes(self):
        c = Crawler()
        with ProcessPoolExecutor(max_workers=4) as executor:
            results = executor.map(c.open, self.urls_list)
        self.assertEqual(len(list(results)), 10)

    def test_run_crawler_in_threads_download_images(self):

        def open_and_download(url):
            response = c.open(url)
            full_images_urls = [c.join_url(src) for src in c.images()]
            downloaded_files = c.download_files(self.test_dir, files=full_images_urls)
            self.assertEqual(len(full_images_urls), len(downloaded_files))
            return response

        c = Crawler()
        with ThreadPoolExecutor(max_workers=4) as executor:
            results = executor.map(open_and_download, self.urls_list)

        self.assertEqual(len(list(results)), 10)

    def test_crawler_custom_submit(self):
        c = Crawler()
        data = {
            'name': 'Luciano Ramalho',
            'title': 'Fluent Python'
        }
        c.submit(
            action=self.urls['POST'],
            data=data
        )
        self.assertEqual(c.response().json().get('form'), data)

    def test_crawler_open_retries(self):
        c = Crawler()
        c.max_retries = 3
        c.logging = True
        with self.assertRaises(ConnectionError):
            c.open('http://www.delver.cg/404', data={'test': 'test data'})
        self.assertEqual(c._retries, c.max_retries)

    def test_crawler_random_timeout(self):
        urls = [
            'https://httpbin.org/html',
            'https://www.w3schools.com/html/html_tables.asp',
            'https://httpbin.org/user-agent'
        ]
        c = Crawler()
        c.random_timeout = (0, 5)
        c.logging = True
        for url in urls:
            c.open(url)


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-

from urllib.parse import urlparse

from lxml import html
from lxml.html.clean import Cleaner

from .forms import FormWrapper
from .helpers import (
    match_form,
    filter_element
)


class HtmlParser:
    """ Parses response content string to valid html using `lxml.html`
    """

    def __init__(self, response, session=None, use_cleaner=None, cleaner_params=None):
        self._html_tree = html.fromstring(response.content)
        self.links = {}
        self._forms = []
        self._cleaner = Cleaner(**cleaner_params) if use_cleaner else None
        self._session = session
        self._url = response.url

    def make_links_absolute(self):
        """Makes absolute links http://domain.com/index.html from the relative ones /index.html
        """
        parsed_url = urlparse(self._url)
        self._html_tree.make_links_absolute(
            '{url.scheme}://{url.netloc}/'.format(url=parsed_url),
            resolve_base_href=True
        )

    def find_links(self, tags=None, filters=None, match='EQUAL'):
        """ Find links and iterate through them checking if they are matching given filters and
        tags

        usage::

        >>> import requests
        >>> response = requests.get('https://httpbin.org/links/10/0')
        >>> tags = ['style', 'link', 'script', 'a']
        >>> parser = HtmlParser(response)
        >>> links = parser.find_links(tags)
        >>> len(links)
        9
        """
        filters = filters or {}
        tags = tags or ['a']
        for link, _, url, _ in self._html_tree.iterlinks():
            matched = filter_element(
                link,
                tags=tags,
                filters=filters,
                match=match
            )
            if matched:
                self.links[url] = matched
        return self.links

    def find_forms(self, filters=None):
        """ Find forms and wraps them with class::`<FormWrapper>` object

        usage::

        >>> import requests
        >>> response = requests.get('https://httpbin.org/forms/post')
        >>> parser = HtmlParser(response)
        >>> forms = parser.find_forms()
        >>> len(forms)
        1
        """
        filters = filters or {}
        self._forms = []
        for form in self._html_tree.forms:
            wrapped_form = FormWrapper(form, session=self._session, url=self._url)
            if match_form(wrapped_form, filters):
                self._forms.append(wrapped_form)
        return self._forms

    def xpath(self, path):
        """Select elements using xpath selectors"""
        return self._html_tree.xpath(path)

    def css(self, selector):
        """Select elements by css selectors"""
        return self._html_tree.cssselect(selector)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding:utf-8 -*-


class GeneralError(Exception):
    """General error"""


class SessionError(GeneralError):
    """Raised on errors related to session usage."""


class HistoryError(GeneralError):
    """Raised when functionality needs history=True."""


class FlowError(GeneralError):
    """Raised on errors related to flow usage."""


class CrawlerError(GeneralError):
    """Raised on errors related to crawler usage."""



# -*- coding: utf-8 -*-

from collections import defaultdict
from lxml.html.clean import Cleaner

__all__ = ['match_form', 'table_to_dict', 'filter_element']

MATCHINGS = {
    'IN': lambda value1, value2: value1 in value2,
    'NOT_IN': lambda value1, value2: value1 not in value2,
    'EQUAL': lambda value1, value2: value1 == value2,
    'NOT_EQUAL': lambda value1, value2: value1 != value2
}


def filter_element(element, tags=None, filters=None, match='EQUAL', custom_attrs=None):
    custom_attrs = custom_attrs or []
    if not tags or element.tag in tags:
        data = {
            'id': element.attrib.get('id', ''),
            'text': element.text,
            'title': element.attrib.get('title', ''),
            'class': element.attrib.get('class', '')
        }
        data.update({attr: element.attrib.get(attr, '') for attr in custom_attrs})

        if filters:
            if match_dict(data, filters, match=match):
                return data
        else:
            return data


def match_dict(data, filters, match='EQUAL'):
    """Matches filters values with data values using predefined matching types

    :param data: dict
    :param filters: dict
    :param match: string, one of  ``'IN', 'NOT_IN', 'EQUAL', 'NOT_EQUAL'``
    :return: bool

    Usage::

    >>> filters = {'class': 'class-120'}
    >>> data = {
    ...     'id': 'searchbox',
    ...     'class': '.what-anice-class-120'
    ... }
    >>> match_dict(data, filters, 'IN')
    True

    """
    for _filter, value in filters.items():
        filtered_data = data.get(_filter)
        if value and filtered_data and MATCHINGS[match](value, filtered_data):
            continue
        else:
            return False
    return True


def match_form(wrapped_form, filters):
    """Matches filters values with <FormElement> attributes using predefined matching types

    example_filters = {
        'id': 'searchbox',
        'name': 'name,
        'action': 'action',
        'has_fields': ['field1', 'field2'],
    }

    :param wrapped_form: class::`FormElement <FormElement>` object
    :param filters: dict
    :return: bool
    """
    for name, value in filters.items():
        if name == 'has_fields':
            if not getattr(wrapped_form, name)(value):
                return False
        elif getattr(wrapped_form, name)() != value:
            return False
    return True


def table_to_dict(table):
    """Turns lxml //table element to dict. Works only with simple flat tables.

    todo:: add handling rowspan [[] []]

    :param table: lxml `<Element>` object
    :return: defaultdict
    """
    cleaner = Cleaner(
        javascript=False,
        style=False
    )

    def process_row():
        for subindex, row_child in enumerate(child.iterchildren()):
            if index == 0:
                row_child = cleaner.clean_html(row_child)
                colspan = int(row_child.attrib.get('colspan') or 0)
                if colspan:
                    row_content = row_child.text_content()
                    colspans[row_content] = colspan
                    for _ in range(colspan):
                        headers.append(row_content)
                else:
                    headers.append(row_child.text_content())
            elif row_child.tag == 'td':
                if headers[subindex] in colspans:
                    if not headers[subindex] in table_dict[index]:
                        table_dict[index][headers[subindex]] = [row_child.text_content()]
                    else:
                        table_dict[index][headers[subindex]].append(row_child.text_content())
                else:
                    table_dict[index][headers[subindex]] = row_child.text_content()
            elif not headers:
                return {}

    headers = []
    colspans = {}
    table_dict = defaultdict(dict)
    for index, child in enumerate(table.iterchildren()):
        if child.tag == 'tr':
            process_row()
    return table_dict


def typed_property(name, expected_type):
    """Common function used to creating arguments with forced type

    :param name: name of attribute
    :param expected_type: expected type of attribute value
    :return: property attribute
    """
    private_name = '_' + name

    @property
    def prop(self):
        return getattr(self, private_name)

    @prop.setter
    def prop(self, value):
        if not isinstance(value, expected_type):
            raise TypeError('Expected {}'.format(expected_type))
        setattr(self, private_name, value)

    return prop


def custom_property(name, type):
    return typed_property(name, type)


# TODO: is it really needed ?
ForcedInteger = lambda name: typed_property(name, int)
ForcedString = lambda name: typed_property(name, str)
ForcedFloat = lambda name: typed_property(name, float)
ForcedBool = lambda name: typed_property(name, bool)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
